@@ -1,7 +1,7 @@
 /*
  * This file is part of GPaste.
  *
- * Copyright (c) 2010-2016, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
+ * Copyright (c) 2010-2017, Marc-Antoine Perennou <Marc-Antoine@Perennou.com>
  */
 
 #include <gpaste-clipboard.h>
@@ -300,6 +300,42 @@ g_paste_clipboard_select_text (GPasteClipboard *self,
     gtk_clipboard_set_text (priv->real, text, -1);
 }
 
+static guchar *
+copy_str_as_uchars (const gchar *str,
+                    guint64      length)
+{
+    guchar *data = g_new (guchar, length);
+    for (guint64 i = 0; i < length; ++i)
+        data[i] = (guchar) str[i];
+    return data;
+}
+
+static void
+_get_clipboard_data_from_special_atom (GtkSelectionData *selection_data,
+                                       const GPasteItem *item,
+                                       GPasteSpecialAtom atom)
+{
+    if (atom >= G_PASTE_SPECIAL_ATOM_FIRST && atom < G_PASTE_SPECIAL_ATOM_LAST)
+    {
+        g_autofree guchar *data = NULL;
+        guint64 length = 0;
+        const gchar *str = g_paste_item_get_special_value (item, atom);
+
+        if (str)
+        {
+            data = g_base64_decode (str, &length);
+        }
+        else
+        {
+            str = g_paste_item_get_value (item);
+            length = strlen (str);
+            data = copy_str_as_uchars (str, length);
+        }
+
+        gtk_selection_data_set (selection_data, g_paste_special_atom_get (atom), 8, data, length);
+    }
+}
+
 static void
 g_paste_clipboard_get_clipboard_data (GtkClipboard     *clipboard G_GNUC_UNUSED,
                                       GtkSelectionData *selection_data,
@@ -310,42 +346,38 @@ g_paste_clipboard_get_clipboard_data (GtkClipboard     *clipboard G_GNUC_UNUSED,
 
     GPasteItem *item = G_PASTE_ITEM (user_data_or_owner);
 
-    GdkAtom targets[1] = { gtk_selection_data_get_target (selection_data) };
+    GdkAtom target = gtk_selection_data_get_target (selection_data);
+    GdkAtom targets[1] = { target };
+
+    if (_G_PASTE_IS_IMAGE_ITEM (item))
+    {
+        if (gtk_targets_include_image (targets, 1, TRUE))
+            gtk_selection_data_set_pixbuf (selection_data, g_paste_image_item_get_image (G_PASTE_IMAGE_ITEM (item)));
+        return;
+    }
+    else if (_G_PASTE_IS_URIS_ITEM (item))
+    {
+        if (gtk_targets_include_uri (targets, 1))
+        {
+            const gchar * const *uris = g_paste_uris_item_get_uris (G_PASTE_URIS_ITEM (item));
+
+            gtk_selection_data_set_uris (selection_data, (GStrv) uris);
+            return;
+        }
+    }
+
+    for (GPasteSpecialAtom a = G_PASTE_SPECIAL_ATOM_FIRST; a < G_PASTE_SPECIAL_ATOM_LAST; ++a)
+    {
+        if (target == g_paste_special_atom_get (a))
+        {
+            _get_clipboard_data_from_special_atom (selection_data, item, a);
+            return;
+        }
+    }
 
     /* The content is requested as text */
     if (gtk_targets_include_text (targets, 1))
         gtk_selection_data_set_text (selection_data, g_paste_item_get_real_value (item), -1);
-    else if (_G_PASTE_IS_IMAGE_ITEM (item))
-    {
-        if (gtk_targets_include_image (targets, 1, TRUE))
-            gtk_selection_data_set_pixbuf (selection_data, g_paste_image_item_get_image (G_PASTE_IMAGE_ITEM (item)));
-    }
-    /* The content is requested as uris */
-    else
-    {
-        g_return_if_fail (_G_PASTE_IS_URIS_ITEM (item));
-
-        const gchar * const *uris = g_paste_uris_item_get_uris (G_PASTE_URIS_ITEM (item));
-
-        if (gtk_targets_include_uri (targets, 1))
-            gtk_selection_data_set_uris (selection_data, (GStrv) uris);
-        /* The content is requested as special gnome-copied-files by nautilus */
-        else
-        {
-            g_autoptr (GString) copy_string = g_string_new ("copy");
-            guint64 length = g_strv_length ((GStrv) uris);
-
-            for (guint64 i = 0; i < length; ++i)
-                g_string_append_printf (copy_string, "\n%s", uris[i]);
-
-            gchar *str = copy_string->str;
-            length = copy_string->len + 1;
-            g_autofree guchar *copy_files_data = g_new (guchar, length);
-            for (guint64 i = 0; i < length; ++i)
-                copy_files_data[i] = (guchar) str[i];
-            gtk_selection_data_set (selection_data, g_paste_clipboard_copy_files_target, 8, copy_files_data, length);
-        }
-    }
 }
 
 static void
@@ -355,32 +387,6 @@ g_paste_clipboard_clear_clipboard_data (GtkClipboard *clipboard G_GNUC_UNUSED,
     g_object_unref (user_data_or_owner);
 }
 
-static void
-g_paste_clipboard_private_select_uris (GPasteClipboardPrivate *priv,
-                                       GPasteUrisItem         *item)
-{
-    GtkClipboard *real = priv->real;
-    g_autoptr (GtkTargetList) target_list = gtk_target_list_new (NULL, 0);
-
-    g_debug("%s: select uris", _g_paste_clipboard_private_target_name (priv));
-
-    g_paste_clipboard_private_set_text (priv, g_paste_item_get_real_value (_G_PASTE_ITEM (item)));
-
-    gtk_target_list_add_text_targets (target_list, 0);
-    gtk_target_list_add_uri_targets (target_list, 0);
-    gtk_target_list_add (target_list, g_paste_clipboard_copy_files_target, 0, 0);
-
-    gint32 n_targets;
-    GtkTargetEntry *targets = gtk_target_table_new_from_list (target_list, &n_targets);
-    gtk_clipboard_set_with_owner (real,
-                                  targets,
-                                  n_targets,
-                                  g_paste_clipboard_get_clipboard_data,
-                                  g_paste_clipboard_clear_clipboard_data,
-                                  g_object_ref (item));
-    gtk_target_table_free (targets, n_targets);
-}
-
 /**
  * g_paste_clipboard_clear:
  * @self: a #GPasteClipboard instance
@@ -554,35 +560,40 @@ g_paste_clipboard_select_item (GPasteClipboard *self,
     g_return_val_if_fail (_G_PASTE_IS_ITEM (item), FALSE);
 
     GPasteClipboardPrivate *priv = g_paste_clipboard_get_instance_private (self);
+    GtkClipboard *real = priv->real;
+    g_autoptr (GtkTargetList) target_list = gtk_target_list_new (NULL, 0);
 
     g_debug("%s: select item", _g_paste_clipboard_private_target_name (priv));
 
     if (_G_PASTE_IS_IMAGE_ITEM (item))
     {
-        const GPasteImageItem *image_item = _G_PASTE_IMAGE_ITEM (item);
-        const gchar *checksum = g_paste_image_item_get_checksum (image_item);
-        GdkPixbuf *image = g_paste_image_item_get_image (image_item);
-
-        if (!image || !GDK_IS_PIXBUF (image))
-            return FALSE;
-        if (!g_paste_str_equal (checksum, priv->image_checksum))
-            g_paste_clipboard_private_select_image (priv, image, checksum);
+        gtk_target_list_add_image_targets (target_list, 0, FALSE);
+        g_paste_clipboard_private_set_image_checksum (priv, g_paste_image_item_get_checksum (G_PASTE_IMAGE_ITEM (item)));
     }
     else
     {
-        const gchar *text = g_paste_item_get_real_value (item);
-
-        if (!g_paste_str_equal (text, priv->text))
-        {
-            if (_G_PASTE_IS_URIS_ITEM (item))
-                g_paste_clipboard_private_select_uris (priv, G_PASTE_URIS_ITEM (item));
-            else  if (_G_PASTE_IS_TEXT_ITEM (item))
-                g_paste_clipboard_select_text (self, text);
-            else
-                g_assert_not_reached ();
-        }
+        gtk_target_list_add_text_targets (target_list, 0);
+        if (_G_PASTE_IS_URIS_ITEM (item))
+            gtk_target_list_add_uri_targets (target_list, 0);
+        g_paste_clipboard_private_set_text (priv, g_paste_item_get_real_value (item));
     }
 
+    for (const GSList *sv = g_paste_item_get_special_values (item); sv; sv = sv->next)
+    {
+        const GPasteSpecialValue *v = sv->data;
+        gtk_target_list_add (target_list, g_paste_special_atom_get (v->mime), 0, 0);
+    }
+
+    gint32 n_targets;
+    GtkTargetEntry *targets = gtk_target_table_new_from_list (target_list, &n_targets);
+    gtk_clipboard_set_with_owner (real,
+                                  targets,
+                                  n_targets,
+                                  g_paste_clipboard_get_clipboard_data,
+                                  g_paste_clipboard_clear_clipboard_data,
+                                  g_object_ref (item));
+    gtk_target_table_free (targets, n_targets);
+
     return TRUE;
 }
 
